module Numeric.BLAS.Matrix.Transposable (
   Transp.Transposable(..), Transp.nonTransposed, Transp.transposed,
   shape, height, width,
   Transp.transpose,
   toMatrix,
   multiplyVector,
   Transp.multiply,
   dotRowwise,
   innerRowwise,
   Transp.kronecker,
   Transp.kroneckerLeft,
   conjugate,
   ) where

import qualified Numeric.BLAS.Matrix.TransposablePrivate as Transp
import qualified Numeric.BLAS.Matrix.RowMajor as Matrix
import qualified Numeric.BLAS.Vector as Vector
import qualified Numeric.Netlib.Class as Class
import Numeric.BLAS.Matrix.TransposablePrivate (Transposable(..))
import Numeric.BLAS.Vector (Vector)

import qualified Data.Array.Comfort.Storable as Array
import qualified Data.Array.Comfort.Shape as Shape



height :: Transposable height width a -> height
height (NonTransposed a) = fst $ Array.shape a
height (Transposed    a) = snd $ Array.shape a

width :: Transposable height width a -> width
width (NonTransposed a) = snd $ Array.shape a
width (Transposed    a) = fst $ Array.shape a

shape :: Transposable height width a -> (height,width)
shape a = (height a, width a)


toMatrix ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Transp.Transposable height width a -> Matrix.Matrix height width a
toMatrix (NonTransposed a) = a
toMatrix (Transposed a) = Matrix.transpose a


multiplyVector ::
   (Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   Transp.Transposable height width a -> Vector width a -> Vector height a
multiplyVector a x = Transp.multiplyVector x (Transp.transpose a)

conjugate ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Transp.Transposable height width a ->
   Transp.Transposable height width a
conjugate (NonTransposed a) = NonTransposed $ Vector.conjugate a
conjugate (Transposed a) = Transposed $ Vector.conjugate a


dotRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
dotRowwise at bt =
   case (at,bt) of
      (Transposed a, Transposed b) -> Matrix.sumColumns $ Vector.mul a b
      _ -> Matrix.sumRows $ Vector.mul (toMatrix at) (toMatrix bt)

innerRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
innerRowwise at bt =
   case (at,bt) of
      (Transposed a, Transposed b) -> Matrix.sumColumns $ Vector.mulConj a b
      _ -> Matrix.sumRows $ Vector.mulConj (toMatrix at) (toMatrix bt)
