-- ToDo: rename to Vector.View
{-# LANGUAGE TypeFamilies #-}
module Numeric.BLAS.Vector.SlicePrivate where

import qualified Numeric.BLAS.Subobject.Shape as Subshape
import qualified Numeric.BLAS.Subobject.View as View
import qualified Numeric.BLAS.Subobject.Layout as Layout
import qualified Numeric.BLAS.Slice as Slice
import qualified Numeric.BLAS.FFI.Generic as Blas
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class

import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Shape as Shape
import qualified Data.StorableVector.Base as SVB
import Data.Array.Comfort.Storable.Unchecked (Array(Array))

import qualified Foreign.Marshal.Array.Guarded as ForeignArray
import Foreign.Marshal.Array (advancePtr)
import Foreign.ForeignPtr (withForeignPtr)
import Foreign.Ptr (Ptr)
import Foreign.Storable (Storable)
import Foreign.C.Types (CInt)

import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)
import Control.Applicative (liftA2)


data Sourced lay sh slice a = Sourced (View.T lay slice) (Array sh a)
   deriving (Show)

newtype T lay slice a = Cons (Array (Subshape.T lay slice) a)
   deriving (Show)


class C v where
   shape :: v slice a -> slice
   mapShape :: (slice0 -> slice1) -> v slice0 a -> v slice1 a
   increment :: v slice a -> Int
   startArg :: (Storable a) => v slice a -> Call.FortranIO r (Ptr a)

instance C Array where
   shape = Array.shape
   mapShape = Array.mapShape
   increment _arr = 1
   startArg (Array _sh x) = ContT $ withForeignPtr x

instance (View.Slice lay) => C (Sourced lay sh) where
   shape (Sourced (Slice.Cons _s _k slc) _arr) = slc
   -- mapShape f (Cons slc arr) = Cons (Slice.mapShape f slc) arr
   mapShape f (Sourced (Slice.Cons s k slc) arr) =
      Sourced (Slice.Cons s k (f slc)) arr
   increment (Sourced (Slice.Cons _s k _slc) _arr) = View.elemInc k
   startArg (Sourced (Slice.Cons s _k _slice) (Array _sh x)) = do
      sxPtr <- ContT $ withForeignPtr x
      return (advancePtr sxPtr s)

instance (View.Slice lay) => C (T lay) where
   shape (Cons arr) = Subshape.shape $ Array.shape arr
   mapShape f (Cons arr) =
      Cons $
--      Array.mapShape (Subshape.mapShape f) arr
      Array.mapShape (\sub -> sub{Subshape.shape = f $ Subshape.shape sub}) arr
   increment (Cons arr) = View.elemInc $ Subshape.layout $ Array.shape arr
   startArg (Cons (Array sh x)) = do
      xPtr <- ContT $ withForeignPtr x
      return $ advancePtr xPtr (Subshape.start sh)


class (C v) => Mono v where
   fromAny :: (Shape.C slice) => v slice a -> T Layout.Slice slice a

instance Mono Array where
   fromAny = Cons . Array.mapShape Subshape.sliceFromVector

instance (View.Slice lay, Shape.C sh) => Mono (Sourced lay sh) where
   fromAny (Sourced (Slice.Cons s lay slc) arr) =
      Cons $
      Array.reshape
         (Subshape.Cons {
            Subshape.start = s,
            Subshape.size = Shape.size $ Array.shape arr,
            Subshape.layout = Slice.sliceFromLayout lay,
            Subshape.shape = slc
         })
         arr

instance (View.Slice lay) => Mono (T lay) where
   fromAny (Cons arr) =
      Cons $
      Array.mapShape
         (\sub -> sub{
            Subshape.layout = Slice.sliceFromLayout $ Subshape.layout sub
         })
         arr


class C v => Indexed v where
   access ::
      (Shape.Indexed sh, Storable a) =>
      v sh a -> Shape.Index sh -> a

instance Indexed Array where
   access = (Array.!)

instance (View.Slice lay, Shape.C shA) => Indexed (Sourced lay shA) where
   access (Sourced (Slice.Cons s k ssh) (Array sh x)) ix =
      Array (Shape.Deferred sh) x
      Array.!
      Shape.DeferredIndex (s + View.elemInc k * Shape.offset ssh ix)

instance (View.Slice lay) => Indexed (T lay) where
   access (Cons (Array sh x)) ix =
      Array (Shape.Deferred sh) x
      Array.!
      Shape.DeferredIndex
         (Subshape.start sh +
            View.elemInc (Subshape.layout sh) *
            Shape.offset (Subshape.shape sh) ix)


class (C v) => Subvector v where
instance Subvector Array where
instance (Layout.Subvector ~ lay) => Subvector (Sourced lay sh) where
instance (Layout.Subvector ~ lay) => Subvector (T lay) where


sliceArg ::
   (C v, Storable a) =>
   v slice a -> Call.FortranIO r (Ptr a, Ptr CInt)
sliceArg x =
   liftA2 (,) (startArg x) (Call.cint $ increment x)

sizeSliceArg ::
   (C v, Shape.C sh, Storable a) =>
   v sh a -> Call.FortranIO r (Ptr CInt, Ptr a, Ptr CInt)
sizeSliceArg x =
   liftA2
      (\nPtr (xPtr,incxPtr) -> (nPtr, xPtr,incxPtr))
      (Call.cint $ Shape.size $ shape x)
      (sliceArg x)


{- |
Like sliceArg but generates and fills a temporary array
in case the BLAS routine forbids increment zero.
-}
sliceFillArg ::
   (C v, Shape.C slice, Class.Floating a) =>
   v slice a -> Call.FortranIO r (Ptr a, Ptr CInt)
sliceFillArg x =
   if increment x == 0
   then do
      let n = Shape.size $ shape x
      nPtr <- Call.cint n
      xPtr <- startArg x
      incxPtr <- Call.cint 0
      yPtr <- ContT $ ForeignArray.alloca n
      incyPtr <- Call.cint 1
      liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr
      return (yPtr, incyPtr)
   else sliceArg x



type Vector = Array

type ShapeInt = Shape.ZeroBased Int


infixl 4 <*|>

(<*|>) ::
   (C v, Storable a) =>
   Call.FortranIO r (Ptr a -> Ptr CInt -> b) ->
   v slice a ->
   Call.FortranIO r b
f <*|> x  =  fmap uncurry f <*> sliceArg x


-- ToDo: shall we re-use the original array in case of no need to copy?
toVector ::
   (C v, Shape.C slice, Class.Floating a) =>
   v slice a -> Vector slice a
toVector x =
   Array.unsafeCreateWithSize (shape x) $ \n yPtr ->
   evalContT $ Call.run $
      pure Blas.copy
         <*> Call.cint n
         <*|> x
         <*> pure yPtr
         <*> Call.cint 1

{- |
Non-copying conversion from @StorableVector@.
-}
fromStorableVector ::
   (View.Slice lay) => SVB.Vector a -> T lay ShapeInt a
fromStorableVector xs =
   case SVB.toForeignPtr xs of
      (buffer, start, len) ->
         Cons
            (Array
               (Subshape.Cons {
                  Subshape.start = start,
                  Subshape.size = start+len,
                  Subshape.layout = View.defltLayout,
                  Subshape.shape = Shape.ZeroBased len
               })
               buffer)

sourcedFromStorableVector ::
   (View.Slice lay) => SVB.Vector a -> Sourced lay ShapeInt ShapeInt a
sourcedFromStorableVector xs =
   case SVB.toForeignPtr xs of
      (buffer, start, len) ->
         Sourced
            (Slice.Cons start View.defltLayout (Shape.ZeroBased len))
            (Array (Shape.ZeroBased (start+len)) buffer)

fromVector :: (View.Slice lay, Shape.C sh) => Vector sh a -> T lay sh a
fromVector xs = Cons $ Array.mapShape Subshape.fromVector xs

-- ToDo: Can we unify some of those functions using C class?
slice :: (View.T lay shA -> View.T lay shB) -> T lay shA a -> T lay shB a
slice f (Cons xs) = Cons (Array.mapShape (Subshape.focus f) xs)

sourceSlice :: (View.T lay shA -> View.T lay shB) -> Sourced lay sh shA a -> Sourced lay sh shB a
sourceSlice f (Sourced slc xs) = Sourced (f slc) xs

sliceVector :: (Shape.C shA) =>
   (Slice.T shA -> Slice.T shB) ->
   Vector shA a -> T Layout.Slice shB a
sliceVector f = slice f . fromVector

viewVector :: (View.Slice lay, Shape.C shA) =>
   (View.T lay shA -> View.T lay shB) ->
   Vector shA a -> T lay shB a
viewVector f = slice f . fromVector

extract ::
   (Shape.C slice, Shape.C sh, Class.Floating a) =>
   (Slice.T sh -> Slice.T slice) -> Vector sh a -> Vector slice a
extract slc xs = toVector $ sliceVector slc xs
